﻿using System;
using System.Linq;
using System.IO.Ports;
using System.Threading.Tasks;
using System.Runtime.InteropServices;
using System.Security.Policy;


/* KeiganMotor API
 * 
 * Send Command to KeiganMotor
 * 
 */

public partial class KeiganMotor_USB
{

    public void TXBuffer(byte[] param,int size)
    {
        Console.WriteLine("TX: " + BitConverter.ToString(param));
        serialPort.Write(param, 0, size);
    }

    public byte[] ReadPIDParams(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x42; // readPIDParams コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;


        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    //public byte[] ReadDeviceName()
    //{
    //    byte[] param = new byte[5];
    //    int size = 0;

    //    //パラメータのセット
    //    param[size++] = 0x46; // readDeviceName コマンド

    //    GetIDBytes().CopyTo(param, size);
    //    size += 2;


    //    GetCRC16Bytes(param, size).CopyTo(param, size);
    //    size += 2;

    //    if (serialPort.IsOpen == true)
    //    {
    //        TXBuffer(param, size);
    //    }
    //    return param;
    //}

    //public byte[] ReadDeviceInfo()
    //{
    //    byte[] param = new byte[5];
    //    int size = 0;

    //    //パラメータのセット
    //    param[size++] = 0x47; // readDeviceInfo コマンド

    //    GetIDBytes().CopyTo(param, size);
    //    size += 2;


    //    GetCRC16Bytes(param, size).CopyTo(param, size);
    //    size += 2;

    //    if (serialPort.IsOpen == true)
    //    {
    //        TXBuffer(param, size);
    //    }
    //    return param;
    //}
    public byte[] ReadAllInfo(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x49; // readAllInfo コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;


        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


  
    // Registers

    public byte[] CurveType(byte type, UInt16 id = 0)
    {
        byte[] param = new byte[6];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x05;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = type;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] Acc(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        byte[] array = FloatToBytes(value);

        //パラメータのセット
        param[size++] = 0x07;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] Dec(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        byte[] array = FloatToBytes(value);

        //パラメータのセット
        param[size++] = 0x08;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    public byte[] MaxTorque(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        byte[] array = FloatToBytes(value);

        //パラメータのセット
        param[size++] = 0x0E;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] CurrentQPIDP(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] array = FloatToBytes(value);

        //パラメータのセット
        param[size++] = 0x18;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] CurrentQPIDI(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] array = FloatToBytes(value);

        //パラメータのセット
        param[size++] = 0x19;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }



    public byte[] CurrentQPIDD(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] array = FloatToBytes(value);

        //パラメータのセット
        param[size++] = 0x1A;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    public byte[] SpeedPIDP(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] array = FloatToBytes(value);

        //パラメータのセット
        param[size++] = 0x1B;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] SpeedPIDI(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] array = FloatToBytes(value);

        //パラメータのセット
        param[size++] = 0x1C;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }



    public byte[] SpeedPIDD(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] array = FloatToBytes(value);

        //パラメータのセット
        param[size++] = 0x1D;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] PositionPIDP(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] array = FloatToBytes(value);


        //パラメータのセット
        param[size++] = 0x1E; //positionPIDP コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    public byte[] PositionPIDI(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] array = FloatToBytes(value);


        //パラメータのセット
        param[size++] = 0x1F; //positionPIDI コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] PositionPIDD(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] array = FloatToBytes(value);


        //パラメータのセット
        param[size++] = 0x20; //positionPIDD コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] PosPIDThreshold(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] array = FloatToBytes(value);


        //パラメータのセット
        param[size++] = 0x21;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] PosArrivalSetting(bool notifyEnabled, float tolerance, UInt32 judgeTime, UInt16 id = 0)
    {
        byte[] param = new byte[14];
        int size = 0;

        param[size++] = 0x2B;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = notifyEnabled ? (byte)1:(byte)0;

        byte[] array = FloatToBytes(tolerance);
        array.CopyTo(param, size);
        size += 4;

        byte[] array2 = UInt32ToBytes(judgeTime);
        array2.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;

    }



    // Low-pass filter の種類
    public enum LowpassFilterType
    {
        LPFCurrent = 0,
        LPFVelocity = 1,
        LPFPosition = 2
    }

    public byte[] LowPassFilter(LowpassFilterType type, float value, UInt16 id = 0)
    {
        byte[] param = new byte[10];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] array = FloatToBytes(value);


        //パラメータのセット
        param[size++] = 0x27; // LowpassFilter コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = (Byte)type;

        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }



    // PID の種類
    public enum PIDParamType
    {
        PID_PARAM_Q_CURRENT_P = 0,
        PID_PARAM_Q_CURRENT_I = 1,
        PID_PARAM_Q_CURRENT_D = 2,
        PID_PARAM_Q_CURRENT_LPF_INPUT = 10,
        PID_PARAM_Q_CURRENT_INPUT_GAIN = 11,

        PID_PARAM_VELOCITY_P = 20,
        PID_PARAM_VELOCITY_I = 21,
        PID_PARAM_VELOCITY_D = 22,
        PID_PARAM_VELOCITY_LPF_INPUT = 30,

        PID_PARAM_POSITION_P = 40,
        PID_PARAM_POSITION_I = 41,
        PID_PARAM_POSITION_D = 42,
        PID_PARAM_POSITION_LPF_INPUT = 50,
        PID_PARAM_POSITION_ID_THRESHOLD = 55,

    }

    public byte[] SetPIDTableValue(byte tableIndex, PIDParamType type, float value, UInt16 id = 0)
    {
        byte[] param = new byte[11];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x23; // SetPIDTableValue

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = tableIndex;
        param[size++] = (byte)type;

        byte[] array = FloatToBytes(value);
        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] SelectPIDTable(byte tableIndex, UInt16 id = 0)
    {
        byte[] param = new byte[6];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x24;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = tableIndex;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] ReadPIDTable(byte tableIndex, UInt16 id = 0)
    {
        byte[] param = new byte[6];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x25;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = tableIndex;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] SavePIDTable(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x26;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] Interface(bool isBLEEnabled, bool isLinkageEnabled, bool isRadioEnabled, bool isUARTEnabled, bool isI2CEnabled, bool isDigitalIOEnabled, bool isUART2Enabled, bool isButtonEnabled)
    {
        Console.WriteLine("{0},{1},{2},{3},{4},{5},{6}", isBLEEnabled, isLinkageEnabled, isUARTEnabled, isI2CEnabled, isDigitalIOEnabled, isUART2Enabled, isButtonEnabled);
        byte flag = 0;
        if (isBLEEnabled) flag |= 1; Console.WriteLine("interface flag: {0}", flag);
        if (isLinkageEnabled) flag |= (1 << 1); Console.WriteLine("interface flag: {0}", flag);
        if (isRadioEnabled) flag |= (1 << 2); Console.WriteLine("interface flag: {0}", flag);
        if (isUARTEnabled) flag |= (1 << 3); Console.WriteLine("interface flag: {0}", flag);
        if (isI2CEnabled) flag |= (1 << 4); Console.WriteLine("interface flag: {0}", flag);
        if (isDigitalIOEnabled) flag |= (1 << 5); Console.WriteLine("interface flag: {0}", flag);
        if (isUART2Enabled) flag |= (1 << 6); Console.WriteLine("interface flag: {0}", flag);
        if (isButtonEnabled) flag |= (1 << 7);
        Console.WriteLine("interface flag: {0}", flag);

        return Interface(flag);
    }

    public byte[] Interface(InterfaceSetting intf)
    {
        return Interface(intf.isBLEEnabled, intf.isLinkageEnabled, intf.isRadioEnabled, intf.isUARTEnabled, intf.isI2CEnabled, intf.isDigitalIOEnabled, intf.isUART2Enabled, intf.isButtonEnabled);
    }

    public byte[] Interface(byte bitFlag, UInt16 id = 0)
    {
        byte[] param = new byte[6];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x2E; // Interfaceコマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = bitFlag; //ビットフラグ

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] SafetySettings(byte bitFlag, UInt16 id = 0)
    {
        byte[] param = new byte[6];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x32; //SafetySettings コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = bitFlag; //ビットフラグ

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] LimitCurrent(float current, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        param[size++] = 0x33;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        byte[] array = FloatToBytes(current);
        array.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] UARTBaudRate(byte rate, UInt16 id = 0)
    {
        byte[] param = new byte[6];
        int size = 0;

        //パラメータのセット
        param[size++] = 0xC3;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = rate; //ビットフラグ

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] I2CAddress(byte address, UInt16 id = 0)
    {
        byte[] param = new byte[6];
        int size = 0;

        //パラメータのセット
        param[size++] = 0xC0;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = address; //ビットフラグ

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    public byte[] SaveAllRegisters(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        param[size++] = 0x41; //saveAllRegisters

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;


        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        System.Threading.Thread.Sleep(3000);
        return param;
    }

    public byte[] ResetAllRegisters(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        param[size++] = 0x4F;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;


        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        System.Threading.Thread.Sleep(500);
        return param;
    }

    public byte[] EnableIMUMeasuement(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        param[size++] = 0xEA;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;


        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    public byte[] DisableIMUMeasuement(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        param[size++] = 0xEB;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;


        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }



    public byte[] EnterDFU(UInt16 id = 0)
    {
        byte[] param = { 0xFD, 0x00, 0x00, 0x4B, 0x73 };


        if (serialPort.IsOpen == true)
        {
            serialPort.Write(param, 0, 5);
        }
        return param;
    }

    public byte[] Disable(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        param[size++] = 0x50; //disable 

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] Enable(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x51; //enable コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            Console.WriteLine(string.Join(", ", param));
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] DisableActionTask(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        param[size++] = 0x6A; //disable 

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] EnableActionTask(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x6B; //enable コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            Console.WriteLine(string.Join(", ", param));
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] Speed(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] spdArray = FloatToBytes(value);

        //パラメータのセット
        param[size++] = 0x58; //speed コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        spdArray.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] SpeedRpm(float rpm, UInt16 id = 0)
    {
        return Speed(rpm * (float)(2 * Math.PI / 60), id);
    }

    public byte[] PresetPosition(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        byte[] spdArray = FloatToBytes(value);

        param[size++] = 0x5A;

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        spdArray.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] PresetPositionDegree(float degree, UInt16 id = 0)
    {
        return PresetPosition(degree * (float)(Math.PI / 180), id);
    }


    public byte[] RunForward(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        param[size++] = 0x60; //runForward コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] RunReverse(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x61; //runReverse コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        //サーボの書き込み
        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    public byte[] RunAtVelocityRpm(float value, UInt16 id = 0)
    {
        return RunAtVelocity(value * (float)(2*Math.PI / 60), id);
    }

    public byte[] RunAtVelocity(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        byte[] spdArray = FloatToBytes(value);

        param[size++] = 0x62;
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        spdArray.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    public byte[] MoveToPosition(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        byte[] spdArray = FloatToBytes(value);

        param[size++] = 0x66; //moveToPosition コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        spdArray.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }
    public byte[] MoveToPositionWait(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        byte[] spdArray = FloatToBytes(value);

        param[size++] = 0x76; //moveToPosition コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        spdArray.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] MoveToPositionDegree(float degree, UInt16 id = 0)
    {
        return MoveToPosition(degree * (float)(Math.PI / 180), id);
    }

    public byte[] MoveToPositionDegreeWait(float degree, UInt16 id = 0)
    {
        return MoveToPositionWait(degree * (float)(Math.PI / 180), id);
    }

    public byte[] MoveByDistance(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        byte[] spdArray = FloatToBytes(value);

        param[size++] = 0x68; //moveByDistance 
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        spdArray.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] MoveByDistanceWait(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        byte[] spdArray = FloatToBytes(value);

        param[size++] = 0x78; //moveByDistance 
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        spdArray.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] MoveByDistanceDegree(float degree, UInt16 id = 0)
    {
        return MoveByDistance(degree * (float)(Math.PI / 180),id);
    }

    public byte[] MoveByDistanceDegreeWait(float degree, UInt16 id = 0)
    {
        return MoveByDistanceWait(degree * (float)(Math.PI / 180), id);
    }


    public byte[] Free(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x6C; //free コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] Stop(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x6D; //stop コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    public byte[] HoldTorque(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        byte[] spdArray = FloatToBytes(value);

        param[size++] = 0x72;
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        spdArray.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] WaitQueue(UInt32 msec, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] dataArray = UInt32ToBytes(msec);

        //パラメータのセット
        param[size++] = 0x92; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        dataArray.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] WaitQueueUntilInput(byte pin, byte edge = 2, UInt32 timeout = 0, byte sub_pin1 = 0xFF, byte sub_state1 = 0, byte sub_pin2 = 0xFF, byte sub_state2 = 0, byte sub_pin3 = 0xFF, byte sub_state3 = 0, UInt16 id = 0)
    {
        byte[] param = new byte[17];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] dataArray = UInt32ToBytes(timeout);

        //パラメータのセット
        param[size++] = 0x96; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = pin;
        param[size++] = edge;

        dataArray.CopyTo(param, size);
        size += 4;

        param[size++] = sub_pin1;
        param[size++] = sub_state1;
        param[size++] = sub_pin2;
        param[size++] = sub_state2;
        param[size++] = sub_pin3;
        param[size++] = sub_state3;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] WaitQueueUntilRadioRXEvent(UInt32 evt, UInt32 timeout, UInt16 id = 0)
    {
        byte[] param = new byte[13];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x98; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        UInt32ToBytes(evt).CopyTo(param, size);
        size += 4;

        UInt32ToBytes(timeout).CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    public byte[] WaitQueueUntilRadioTXEventResp(UInt32 evt, UInt32 interval, UInt32 timeout, Boolean withResp, UInt16 id = 0)
    {
        byte[] param = new byte[18];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x99; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        UInt32ToBytes(evt).CopyTo(param, size);
        size += 4;

        UInt32ToBytes(interval).CopyTo(param, size);
        size += 4;

        UInt32ToBytes(timeout).CopyTo(param, size);
        size += 4;

        param[size++] = withResp? (byte)1:(byte)0;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    // WriteGPIOOutput is not implemented by KeiganMotor side
    public byte[] WriteGPIOOutput(byte pin0, byte state0, byte pin1, byte state1, byte pin2, byte state2, byte pin3, byte state3, UInt32 pulseWidthMsec, UInt16 id = 0)
    {
        byte[] param = new byte[17];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。pulseWidthMsec
        byte[] dataArray = UInt32ToBytes(pulseWidthMsec);

        //パラメータのセット
        param[size++] = 0xC5; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = pin0;
        param[size++] = state0;
        param[size++] = pin1;
        param[size++] = state1;
        param[size++] = pin2;
        param[size++] = state2;
        param[size++] = pin3;
        param[size++] = state3;

        dataArray.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }



    public byte[] TriggerTasksetSettings(UInt16 index, UInt32 repeating, UInt16 id = 0)
    {
        byte[] param = new byte[11];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] indexArray = UInt16ToBytes(index);
        byte[] repeatArray = UInt32ToBytes(repeating);

        //パラメータのセット
        param[size++] = 0xA8; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        indexArray.CopyTo(param, size);
        size += 2;

        repeatArray.CopyTo(param, size);
        size += 4;


        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;

    }


    public byte[] TriggerMotionSettings(UInt16 index, UInt32 repeating, PlaybackOption option, Boolean auto_start, UInt16 id = 0)
    {
        byte[] param = new byte[13];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] indexArray = UInt16ToBytes(index);
        byte[] repeatArray = UInt32ToBytes(repeating);

        //パラメータのセット
        param[size++] = 0xB2; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        indexArray.CopyTo(param, size);
        size += 2;

        repeatArray.CopyTo(param, size);
        size += 4;

        param[size++] = (byte)option;
        param[size++] = auto_start ? (byte)1 : (byte)0;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] AutoStartSetting(Byte mode, UInt16 id = 0)
    {
        byte[] param = new byte[6];
        int size = 0;

        //パラメータのセット
        param[size++] = 0xBC; // コマンド


        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = mode;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] ButtonSetting(Byte mode, UInt16 id = 0)
    {
        byte[] param = new byte[6];
        int size = 0;

        //パラメータのセット
        param[size++] = 0xBD; // コマンド


        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = mode;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    /*
     * Teaching Playback 
     * 
     */

    public byte[] EraseMotion(UInt16 index, UInt16 id = 0)
    {
        byte[] param = new byte[7];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] indexArray = UInt16ToBytes(index);

        //パラメータのセット
        param[size++] = 0xAD; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        indexArray.CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] EraseAllMotion(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0xAE; // コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    public byte[] PrepareTeachingMotion(UInt16 index, UInt32 time, UInt16 id = 0)
    {
        byte[] param = new byte[11];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] indexArray = UInt16ToBytes(index);
        byte[] timeArray = UInt32ToBytes(time);

        //パラメータのセット
        param[size++] = 0xAA; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        indexArray.CopyTo(param, size);
        size += 2;

        timeArray.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] StartTeachingMotionFromPrep(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0xAB; // コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] StopTeachingMotion(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0xAC; // コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] ReadMotion(UInt16 index, UInt16 id = 0)
    {
        byte[] param = new byte[7];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] indexArray = UInt16ToBytes(index);

        //パラメータのセット
        param[size++] = 0xB7; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        indexArray.CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] WriteMotionPosition(float value, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        byte[] spdArray = FloatToBytes(value);

        param[size++] = 0xB8; //moveByDistance 
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        spdArray.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] StartDoingTaskset(UInt16 index, UInt32 repeating, UInt16 id = 0)
    {
        byte[] param = new byte[11];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] indexArray = UInt16ToBytes(index);
        byte[] repeatArray = UInt32ToBytes(repeating);

        //パラメータのセット
        param[size++] = 0x81; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        indexArray.CopyTo(param, size);
        size += 2;

        repeatArray.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] StopDoingTaskset(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x82; // コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    // playback 開始地点の列挙子
    public enum PlaybackOption
    {
        StartFromRawRepeatWithBack = 0,
        StartFromCurrentRepeatWithBack = 1,
        StartFromRawRepeatWithoutBack = 2,
        StartFromCurrentRepeatWithoutBack = 3
    }

    public byte[] StartPlaybackMotion(UInt16 index, UInt32 repeating, PlaybackOption option, UInt16 id = 0)
    {
        byte[] param = new byte[12];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] indexArray = UInt16ToBytes(index);
        byte[] repeatArray = UInt32ToBytes(repeating);

        //パラメータのセット
        param[size++] = 0x85; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        indexArray.CopyTo(param, size);
        size += 2;

        repeatArray.CopyTo(param, size);
        size += 4;

        param[size++] = (byte)option;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] PreparePlaybackMotion(UInt16 index, UInt32 repeating, PlaybackOption option, UInt16 id = 0)
    {
        byte[] param = new byte[12];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] indexArray = UInt16ToBytes(index);
        byte[] repeatArray = UInt32ToBytes(repeating);

        //パラメータのセット
        param[size++] = 0x86; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        indexArray.CopyTo(param, size);
        size += 2;

        repeatArray.CopyTo(param, size);
        size += 4;

        param[size++] = (byte)option;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] StartPlaybackMotionFromPrep(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x87; // コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] StopPlaybackMotion(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x88; // コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] EraseTaskset(UInt16 index, UInt16 id = 0)
    {
        byte[] param = new byte[7];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] indexArray = UInt16ToBytes(index);

        //パラメータのセット
        param[size++] = 0xA3; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        indexArray.CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] EraseAllTasksets(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0xA4; // コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] StartRecordingTaskset(UInt16 index, UInt16 id = 0)
    {
        byte[] param = new byte[7];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] indexArray = UInt16ToBytes(index);

        //パラメータのセット
        param[size++] = 0xA0; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        indexArray.CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }



    public byte[] ReadTasksetInfo(UInt16 index, UInt16 id = 0)
    {
        byte[] param = new byte[7];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] indexArray = UInt16ToBytes(index);

        //パラメータのセット
        param[size++] = 0xA6; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        indexArray.CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        System.Threading.Thread.Sleep(30); // 値取得まで待機
        return param;
    }

    public byte[] ReadTaskset(UInt16 index, UInt16 id = 0)
    {
        byte[] param = new byte[7];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] indexArray = UInt16ToBytes(index);

        //パラメータのセット
        param[size++] = 0xA1; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        indexArray.CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    public byte[] StopRecordingTaskset(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0xA2; // コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] TasksetName(String name, UInt16 id = 0)
    {
        byte[] param = new byte[20];
        int size = 0;

        //パラメータのセット
        param[size++] = 0xA5; // コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        string tasksetName = name;

        byte[] full_data = System.Text.Encoding.UTF8.GetBytes(tasksetName);
        byte[] nameData = full_data.Take(15).ToArray();

        nameData.CopyTo(param, size);

        size += 15;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    /* 
     * Registers
     */

    public byte[] TeachingInterval(UInt32 interval, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] intvlArray = UInt32ToBytes(interval);

        //パラメータのセット
        param[size++] = 0x16; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        intvlArray.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] PlaybackInterval(UInt32 interval, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        // Windows PC　はリトルエンディアンであることに注意。ビッグエンディアンにしてバイト列にする。
        byte[] intvlArray = UInt32ToBytes(interval);

        //パラメータのセット
        param[size++] = 0x17; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        intvlArray.CopyTo(param, size);
        size += 4;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    /*
     *      
        Set the notification interval of motor measurement values. It is valid only for USB(UART) and BLE.
        In case of BLE, the allowed minimum value is 100 [ms]. 
        
        flag: interval[ms]
        -----------
        0:      2 [ms]
        1:      5 [ms]
        2:     10 [ms]
        3:     20 [ms]
        4:     50 [ms]
        5:    100 [ms]
        6:    200 [ms]
        7:    500 [ms]
        8:   1000 [ms]
        9:   2000 [ms]
        10:  5000 [ms]
        11: 10000 [ms]
        -----------        
     */
    public byte[] MotorMeasurementInterval(Byte flag, UInt16 id = 0)
    {
        byte[] param = new byte[6];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x2C; // コマンド


        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = flag;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    /*
     *            
        Set the notification interval of IMU measurement values. It is valid only for USB(UART) and BLE.
        In case of BLE, the allowed minimum value is 100 [ms]. 
        flag: interval[ms]
        -----------
        0:     10 [ms]
        1:     20 [ms]
        2:     50 [ms]
        3:    100 [ms]
        4:    200 [ms]
        5:    500 [ms]
        6:   1000 [ms]
        7:   2000 [ms]
        9:   5000 [ms]
        10: 10000 [ms]
        -----------      
     */
    public byte[] IMUmeasurementInterval(Byte flag, UInt16 id = 0)
    {
        byte[] param = new byte[6];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x3C; // コマンド


        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = flag;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    /*
     * Read  
     *  
     */
    public byte[] ReadMotorMeasurement(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0xB4; // コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] ReadStatus(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x9A; // コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }


    public byte[] ReadRegister(Byte register, UInt16 id = 0)
    {
        byte[] param = new byte[6];
        int size = 0;

        //パラメータのセット
        param[size++] = 0x40; // コマンド

        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = register;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        System.Threading.Thread.Sleep(30); // 値取得まで待機
        return param;
    }


    public InterfaceSetting ReadInterface(UInt16 id = 0)
    {
        ReadRegister(0x2E);
        
        return interfaceSetting;

    }
    

    
    public String ReadDeviceName(UInt16 id = 0)
    {
        ReadRegister(0x46);
        return deviceName;
    }

    public String ReadDeviceInfo(UInt16 id = 0)
    {
        ReadRegister(0x47);
        return deviceInfo;
    }


    public float ReadAcc(UInt16 id = 0)
    {
        ReadRegister(0x07);
        return acc;
    }

    public float ReadDec(UInt16 id = 0)
    {
        ReadRegister(0x08);
        return dec;
    }

    public float ReadSpeedP(UInt16 id = 0)
    {
        ReadRegister(0x1B);
        return speedP;
    }

    public float ReadSpeedI(UInt16 id = 0)
    {
        ReadRegister(0x1C);
        return speedI;
    }

    public float ReadSpeedD(UInt16 id = 0)
    {
        ReadRegister(0x1C);
        return speedD;
    }

    public float ReadPositionP(UInt16 id = 0)
    {
        ReadRegister(0x1E);
        return positionP;
    }

    public float ReadPositionI(UInt16 id = 0)
    {
        ReadRegister(0x1F);
        return positionI;
    }

    public float ReadPositionD(UInt16 id = 0)
    {
        ReadRegister(0x20);
        return positionD;
    }

    public float ReadPosPIDThreshold(UInt16 id = 0)
    {
        ReadRegister(0x21);
        return positionPIDThreshold;
    }

    public (bool, float, UInt32) ReadPosArrivalSetting(UInt16 id = 0)
    {
        ReadRegister(0x2B);
        return (notifyPosArrival, posArrivalTolerance, posArrivalJudgeTime);
    }


    /*
     * RADIO
     */
    // RADIOモードの列挙子
    public enum RadioMode
    {
        NONE = 0, // なにもしない
        EVENT_TX = 1, // イベントモード送信（シンプルなイベント送受信モード KMQuick 想定）
        EVENT_RX = 2, // イベントモード受信（シンプルなイベント送受信モード KMQuick 想定）
        PARENT = 3, // 親モード（コマンドを直接投げる）
        CHILD = 4 // 子供モード（親からのコマンドを直接受け付ける）

    }

    public byte[] RadioSettings(UInt32 key, byte reserved, RadioMode mode, UInt16 id = 0)
    {
        byte[] param = new byte[11];
        int size = 0;

        //パラメータのセット
        param[size++] = 0xDB; 
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;
        UInt32ToBytes(key).CopyTo(param, size);
        size += 4;
        param[size++] = reserved; // not used
        param[size++] = (byte)mode;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            Console.WriteLine("serialPort.IsOpen");
            TXBuffer(param, size);
        }
        else
        {
            Console.WriteLine("serialPort.Is not Open");
        }

        Console.WriteLine(param);
        return param;
    }

    public byte[] RadioSettingsTextKey(string key, byte reserved, RadioMode mode, UInt16 id = 0)
    {
        char[] keyChar = key.ToCharArray();
        byte[] dummy = { 0 };
        if (key.Length != 4) return dummy;
        UInt32 keyValue = (UInt32)key[0] << 24 |
                            (UInt32)key[1] << 16 |
                             (UInt32)key[2] << 8 |
                              (UInt32)key[3] << 0;
        Console.WriteLine("Key String[]: {0}, UInt32 : {1}", key, keyValue);
        return RadioSettings(keyValue, reserved, mode, id);

    }

    public (string, byte, RadioMode) ReadRadioSettingsTextKey(UInt16 id = 0)
    {
        ReadRegister(0xDB);
        return (radioKey, radioGroup, radioMode);
    }

    /*
     * LED  
     *  
     */


    // LED状態の列挙子
    public enum LedState
    {
        Off,
        OnSolid,
        OnFlash,
        OnDim
    }

    public byte[] Led(LedState state, byte red, byte green, byte blue, UInt16 id = 0)
    {
        byte[] param = new byte[9];
        int size = 0;

        //パラメータのセット
        param[size++] = 0xE0; //LED コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;
        param[size++] = (byte)state; // LedState
        param[size++] = red; //RGB
        param[size++] = green; //RGB
        param[size++] = blue; //RGB

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            Console.WriteLine("serialPort.IsOpen");
            TXBuffer(param, size);
        } else
        {
            Console.WriteLine("serialPort.Is not Open");
        }
        Console.WriteLine(param);
        return param;
    }


    /*
     * Others
     */
    public byte[] Reboot(UInt16 id = 0)
    {
        byte[] param = new byte[5];
        int size = 0;

        //パラメータのセット
        param[size++] = 0xF0; // コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }

    public byte[] EnableCheckSum(bool isEnabled, UInt16 id = 0)
    {
        byte[] param = new byte[6];
        int size = 0;

        //パラメータのセット
        param[size++] = 0xF3; // コマンド
        UInt16ToBytes(id).CopyTo(param, size);
        size += 2;

        param[size++] = isEnabled ? (byte)1 : (byte)0;

        GetCRC16Bytes(param, size).CopyTo(param, size);
        size += 2;

        if (serialPort.IsOpen == true)
        {
            TXBuffer(param, size);
        }
        return param;
    }



}
