﻿using System;
using System.Linq;
using System.IO.Ports;


/* KeiganMotor Utility 
 * 
 * - Type Conversion (example: byte[] <==> float, UInt16, UInt32)
 * - Others
 * 
 */

public partial class KeiganMotor_USB
{
    static UInt16[] crc16_table = new UInt16[256];

    enum Endian
    {
        Little, Big
    }

    static byte[] CorrectEndianBytes(byte[] bytes, Endian endian)
    {
        if (BitConverter.IsLittleEndian ^ (endian == Endian.Little))
            return bytes.Reverse().ToArray();
        else
            return bytes;
    }

    static byte[] FloatToBytes(float value)
    {
        return CorrectEndianBytes(BitConverter.GetBytes(value), Endian.Big);

    }

    static byte[] UInt16ToBytes(UInt16 value)
    {
        return CorrectEndianBytes(BitConverter.GetBytes(value), Endian.Big);

    }

    static byte[] UInt16ToBytesLittle(UInt16 value)
    {
        return CorrectEndianBytes(BitConverter.GetBytes(value), Endian.Little);

    }

    static byte[] Int16ToBytes(Int16 value)
    {
        return CorrectEndianBytes(BitConverter.GetBytes(value), Endian.Big);

    }

    static byte[] UInt32ToBytes(UInt32 value)
    {
        return CorrectEndianBytes(BitConverter.GetBytes(value), Endian.Big);

    }



    static float BytesToFloat(byte[] bytes)
    {
        //DEBUG
        //Console.Write("bytes: ");
        //Console.WriteLine(BitConverter.ToString(bytes));

        byte[] correctedBytes = CorrectEndianBytes(bytes, Endian.Big);

        if (bytes.Count() == sizeof(float))
        {
            return BitConverter.ToSingle(correctedBytes, 0);
        }
        else
        {
            return 0;
        }
    }

    static UInt16 BytesToUInt16(byte[] bytes)
    {
        byte[] correctedBytes = bytes.Reverse().ToArray();
        if (bytes.Count() == sizeof(UInt16))
        {
            return BitConverter.ToUInt16(correctedBytes, 0);
        }
        else
        {
            return 0;
        }
    }

    static Int16 BytesToInt16(byte[] bytes)
    {
        byte[] correctedBytes = bytes.Reverse().ToArray();
        if (bytes.Count() == sizeof(Int16))
        {
            return BitConverter.ToInt16(correctedBytes, 0);
        }
        else
        {
            return 0;
        }
    }

    static UInt32 BytesToUInt32(byte[] bytes)
    {
        byte[] correctedBytes = bytes.Reverse().ToArray();
        if (bytes.Count() == sizeof(UInt32))
        {
            return BitConverter.ToUInt32(correctedBytes, 0);
        }
        else
        {
            return 0;
        }
    }

    static void CRC16Init()
    {
        for (UInt16 i = 0; i < 256; i++)
        {
            UInt16 c = i;
            for (int j = 0; j < 8; j++)
            {
                c = ((c & 1) == 1) ? (UInt16)(0x8408 ^ (c >> 1)) : (UInt16)(c >> 1); // CRC16POLY 0x8408 右送り CRC-16-CCITT (x16 + x12 + x5 + 1) ビット反転（初期値=0）
            }
            crc16_table[i] = c;
        }

        //Console.WriteLine(string.Join(", ", crc16_table));
    }

    static UInt16 CRC16(byte[] buf, int len)
    {
        UInt16 c = 0;
        for (UInt32 i = 0; i < len; i++)
        {
            var num = (c ^ buf[i]) & 0xFF;
            c = (UInt16)(crc16_table[num] ^ (c >> 8));
        }
        return c;
    }

    static byte[] GetCRC16Bytes(byte[] buf, int len)
    {
        UInt16 crc_value = CRC16(buf, len);
        return UInt16ToBytesLittle(crc_value); // リトルエンディアンであることに注意
    }

}
